#ifndef _LIBCW_TESTS_GEN_CW_GEN_ENQUEUE_CHARACTER_NO_ICS_H_
#define _LIBCW_TESTS_GEN_CW_GEN_ENQUEUE_CHARACTER_NO_ICS_H_




#include "test_framework.h"
#include "libcw_gen_tests.h"




cwt_retv test_cw_gen_enqueue_character_no_ics(cw_test_executor_t * cte);




#endif /* #ifndef _LIBCW_TESTS_GEN_CW_GEN_ENQUEUE_CHARACTER_NO_ICS_H_ */

