/*
  This file is a part of unixcw project.
  unixcw project is covered by GNU General Public License, version 2 or later.
*/

#ifndef H_LIBCW_NULL
#define H_LIBCW_NULL




#include "libcw2.h"
#include "libcw_gen.h"




cw_ret_t cw_null_init_gen_internal(cw_gen_t * gen);




#endif /* #ifndef H_LIBCW_NULL */
