#ifndef TEST_CW_GET_SEND_PARAMETERS_H
#define TEST_CW_GET_SEND_PARAMETERS_H




#include "test_framework.h"




int legacy_api_test_cw_get_send_parameters(cw_test_executor_t * cte);




#endif /* #ifndef TEST_CW_GET_SEND_PARAMETERS_H */

