#ifndef CWUTILS_TESTS_CMDLINE_COMBINE_ARGUMENTS_H
#define CWUTILS_TESTS_CMDLINE_COMBINE_ARGUMENTS_H




/**
   @brief Tests of cmdline/combine_arguments()

   @reviewedon 2023.09.02

   @return 0 if tests passed
   @return -1 otherwise
*/
int test_combine_arguments(void);




#endif /* #ifndef CWUTILS_TESTS_CMDLINE_COMBINE_ARGUMENTS_H */

