/*
  This file is a part of unixcw project.  unixcw project is covered by
  GNU General Public License, version 2 or later.
*/

#ifndef _LIBCW_LEGACY_TESTS_REC_POLL_H_
#define _LIBCW_LEGACY_TESTS_REC_POLL_H_




#include "test_framework.h"




/* 'Receiver' test area - poll interface as used in xcwcp from package 3.5.1 and earlier. */
cwt_retv legacy_api_test_rec_poll(cw_test_executor_t * cte);




#endif /* #ifndef _LIBCW_LEGACY_TESTS_REC_POLL_H_ */
