#ifndef _LIBCW_TESTS_CW_GEN_GET_TIMING_PARAMETERS_INTERNAL_H_
#define _LIBCW_TESTS_CW_GEN_GET_TIMING_PARAMETERS_INTERNAL_H_




#include "test_framework.h"




cwt_retv test_cw_gen_get_timing_parameters_internal(cw_test_executor_t * cte);




#endif /* #ifndef _LIBCW_TESTS_CW_GEN_GET_TIMING_PARAMETERS_INTERNAL_H_ */

