/*
  This file is a part of unixcw project.  unixcw project is covered by
  GNU General Public License, version 2 or later.
*/

#ifndef _LIBCW_GEN_TESTS_STATE_CALLBACK_H_
#define _LIBCW_GEN_TESTS_STATE_CALLBACK_H_




#include "test_framework.h"




int test_cw_gen_state_callback(cw_test_executor_t * cte);




#endif /* _LIBCW_GEN_TESTS_STATE_CALLBACK_H_ */
