#ifndef _LIBCW_GEN_TESTS_DEBUG_PCM_FILE_TIMINGS_H_
#define _LIBCW_GEN_TESTS_DEBUG_PCM_FILE_TIMINGS_H_




#include "test_framework.h"




int test_cw_gen_debug_pcm_file_timings(cw_test_executor_t * cte);




#endif /* _LIBCW_GEN_TESTS_DEBUG_PCM_FILE_TIMINGS_H_ */

