/*
 * Copyright (C) 2001-2006  Simon Baldwin (simon_baldwin@yahoo.com)
 * Copyright (C) 2011-2023  Kamil Ignacak (acerion@wp.pl)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef H_CW_COPYRIGHT
#define H_CW_COPYRIGHT

#include "i18n.h"

#if defined(__cplusplus)
extern "C" {
#endif

	/* Copyright notice copied from https://www.gnu.org/licenses/old-licenses/gpl-2.0.html#SEC4 */
static const char *const CW_COPYRIGHT =
	N_("Copyright (C) 2001-2006  Simon Baldwin\n"
	   "Copyright (C) 2011-2023  Kamil Ignacak\n"
	   "\n"
	   "This program is free software; you can redistribute it and/or\n"
	   "modify it under the terms of the GNU General Public License\n"
	   "as published by the Free Software Foundation; either version 2\n"
	   "of the License, or (at your option) any later version.\n"
	   "\n"
	   "This program is distributed in the hope that it will be useful,\n"
	   "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
	   "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
	   "GNU General Public License for more details.\n"
	   "\n"
	   "You should have received a copy of the GNU General Public License\n"
	   "along with this program; if not, write to the Free Software\n"
	   "Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.\n");

#if defined(__cplusplus)
}
#endif
#endif  /* H_CW_COPYRIGHT */
