/*
  This file is a part of unixcw project.  unixcw project is covered by
  GNU General Public License, version 2 or later.
*/

#ifndef _LIBCW_TEST_TQ_SHORT_SPACE_H_
#define _LIBCW_TEST_TQ_SHORT_SPACE_H_




#include "test_framework.h"




int legacy_api_test_tq_short_space(cw_test_executor_t * cte);




#endif /* #ifndef _LIBCW_TEST_TQ_SHORT_SPACE_H_ */
